@@ -3,6 +3,7 @@ from elasticsearch_dsl import Search
 from elasticsearch_dsl.response import Response
 from minetext.config import Config
 from minetext.domain.es_request import EsRequest
+from datetime import datetime
 
 
 class Mine:
@@ -15,6 +16,8 @@ class Mine:
     _verification_uri_complete: str
     _access_token: str
     _refresh_token: str
+    _token_creation_time: float
+    _expiration_time: int
 
     def __init__(self, es_request: EsRequest, host: str = Config.host):
         """
@@ -31,12 +34,22 @@ class Mine:
         self._verification_uri_complete = ''
         self._access_token = ''
         self._refresh_token = ''
+        self._token_creation_time = -1
+        self._expiration_time = -1
 
     def search(self) -> Response:
         """
-        Call the search endpoint with parameters provided via the ``_es_request`` property.
+        Call the search endpoint with parameters provided via the :py:attr:`_es_request` property.
 
-        :return: the result wrapped in the ``Response`` object
+        Returns
+        -------
+        result : Response
+            the search result wrapped in the :py:class:`Response` object.
+
+        Raises
+        ------
+        HTTPError
+            if the request failed.
         """
         url = f'{self._host}/document/search'
 
@@ -49,11 +62,28 @@ class Mine:
             'wa': self._es_request.analytics
         }
 
-        if not self._access_token:
+        if self._access_token:
+
+            # Refresh the access token if necessary
+            if self._token_needs_refresh():
+                self._refreshing_token()
+
+            # Use the access token
             headers = {
                 'Authorization': f'Bearer {self._access_token}'
             }
-            result = requests.get(url, params=payload, headers=headers)
+
+            try:
+                result = requests.get(url, params=payload, headers=headers)
+            except requests.HTTPError as e:
+                # This is when the user is unauthorized. If they have an _access_token but 
+                # are unauthorized the _access_token probably expired. 401 is unauthorized.
+                if e.response.status_code == 401:
+                    self._refreshing_token()
+                    headers['Authorization'] = f'Bearer {self._access_token}'
+                    result = requests.get(url, params=payload, headers=headers)
+                else:
+                    raise e
         else:
             result = requests.get(url, params=payload)
 
@@ -62,8 +92,55 @@ class Mine:
 
         return response
 
+    def _token_needs_refresh(self) -> bool:
+        """
+        Checks if the access token is still usable within the next 20 seconds.
+
+        Returns
+        -------
+        True
+            if the token is (almost) expired.
+        False
+            if the token is still usable.
+        """
+        curr_time = datetime.now().timestamp()
+        return curr_time - self._token_creation_time > 1000 * (self._expiration_time - 20)
+
+    def _refreshing_token(self) -> None:
+        """
+        Creates a new access token with the refresh token.
+
+        Sets a new refresh token, access token and updates the expiration times.
+        When the refresh token does not work the user is forwarded to login again.
+
+        Raises
+        -------
+        HTTPError 
+            When the token could not be refreshed.
+        """
+        payload = {
+            'refresh_token': self._refresh_token
+        }
+        try:
+            token_response = requests.post(f"{self.host}/auth/refresh_token", json=payload)
+        except requests.HTTPError:
+            # This should realistically not happen as the refresh token gets updated every time the 
+            # authorization token is updated and the refresh token has a very long expiration time.
+            print('You need to login again.')
+            self.login()
+            return
+        self._set_values(token_response.json())
+
+    def _set_values(self, json) -> None:
+        self._access_token = json['access_token']
+        self._refresh_token = json['refresh_token']
+        self._expiration_time = json['expires_in']
+        self._refresh_expiration_time = json['refresh_expires_in']
+        self._token_creation_time = datetime.now().timestamp()
+
     def _create_device_token(self) -> None:
-        """Sets device token from mine-graph-api.
+        """
+        Sets device token from mine-graph-api.
 
         Sets the uris, device_code and user_code from response.
 
@@ -81,7 +158,8 @@ class Mine:
         self._verification_uri_complete = token_json['verification_uri_complete']
 
     def _create_access_token(self) -> None:
-        """Sets access token from authentication.
+        """
+        Sets access token from authentication.
 
         Posts the device_code to MINE-API. If the user logged in properly
         the device code should now authorize them. Sets the access_token of Auth object.
@@ -94,13 +172,12 @@ class Mine:
         """
         token_response = requests.post(f'{self.host}/auth/token', json={'device_code': self._device_code})
         token_response.raise_for_status()
-        token_json = token_response.json()
-        self._access_token = token_json['access_token']
-        self._refresh_token = token_json['refresh_token']
+        self._set_values(token_response.json())
         print('Login successful! You are now authorized.')
 
     def login(self) -> None:
-        """Calls the functions to authorize user.
+        """
+        Calls the functions to authorize user.
 
         Waits for the user to log in into the verification uri and afterwards creates an access token if the user
         stated they granted access.
